/*
 * Run a program using the supplied arguments.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <errno.h>
#include <fcntl.h>
#include <port.h>
#include <process.h>
#include <stdio.h>
#include <sys/types.h>



/* Transfer size for communication. */

static const offset_t TO_TRANSFER = 1024;

/* Transfer data between streams. */

static void transfer(FILE *input, FILE *output, char *buffer, int *start, size_t *remaining)
{
    size_t nwritten;

    /* Transfer any previously read data before trying to read more. */

    while (*remaining || (*start = 0, *remaining = fread(buffer, sizeof(char), TO_TRANSFER, input)))
    {
        /* Write the data, stopping if no data can be written. */

        nwritten = fwrite(buffer + *start, sizeof(char), *remaining, output);

        if (!nwritten)
            return;

        /* Record the remaining amount. */

        *remaining -= nwritten;
        *start += nwritten;
    }
}



/* List objects in the filesystem image. */

int main(int argc, char *argv[])
{
    FILE *input, *output, *error;
    pid_t pid;
    port_t port;
    event_source_t source;
    event_values_t result;

    /* Transfer buffers and status. */

    char input_buffer[TO_TRANSFER];
    char output_buffer[TO_TRANSFER];
    char error_buffer[TO_TRANSFER];
    int input_start = 0, output_start = 0, error_start = 0;
    size_t input_remaining = 0, output_remaining = 0, error_remaining = 0;

    if (argc < 2)
    {
        printf("Usage: test_popenv <program> <argument>...\n");
        return 1;
    }

    printf("Running: %s\n", argv[1]);

    pid = popenv(argc - 1, (const char **) argv + 1, &input, &output, &error);

    if (pid == -1)
    {
        printf("Error: %d\n", errno);
        return 1;
    }

    printf("Process: %d\n", pid);

    /* Make stdin and the pipes non-blocking to be able to read and write as
       much data as is available and to monitor events on them manually. */

    if (stdin != NULL)
    {
        fcntl(fileno(stdin), F_SETFL, O_NONBLOCK | fcntl(fileno(stdin), F_GETFL));
        fcntl(fileno(input), F_SETFL, O_NONBLOCK | fcntl(fileno(input), F_GETFL));
    }
    else
    {
        fclose(input);
        input = NULL;
    }

    fcntl(fileno(output), F_SETFL, O_NONBLOCK | fcntl(fileno(output), F_GETFL));

    if (error != NULL)
        fcntl(fileno(error), F_SETFL, O_NONBLOCK | fcntl(fileno(error), F_GETFL));

    /* Monitor the streams and the process. */

    port = port_create();

    port_subscribe(port, port_process_source(pid), PORT_PROCESS_ALL);

    if (stdin != NULL)
    {
        port_subscribe(port, port_file_source(fileno(stdin)), PORT_INPUT | PORT_CLOSE);
        port_subscribe(port, port_file_source(fileno(input)), PORT_OUTPUT | PORT_CLOSE);
    }

    port_subscribe(port, port_file_source(fileno(output)), PORT_INPUT | PORT_CLOSE);

    if (error != NULL)
        port_subscribe(port, port_file_source(fileno(error)), PORT_INPUT | PORT_CLOSE);

    /* Wait for events on the streams and the process. */

    while (1)
    {
        if (port_wait(port, &source))
        {
            printf("Error waiting for notifications.\n");
            return 1;
        }

        /* Handle input to this process and its transfer to the created
           process. */

        if ((stdin != NULL) && (input != NULL))
        {
            /* Attempt to transfer input. */

            if ((source == port_file_source(fileno(stdin))) ||
                input_remaining || (source == port_file_source(fileno(input))))

                transfer(stdin, input, input_buffer, &input_start, &input_remaining);

            /* If no input remains and stdin has been closed, close the input
               pipe. */

            if (!input_remaining && (feof(stdin) ||
                 port_source_event_flags(port_file_source(fileno(stdin))) & PORT_CLOSE))
            {
                fclose(input);
                input = NULL;
            }
        }

        /* Handle output from the process. */

        if (output_remaining || (source == port_file_source(fileno(output))))
            transfer(output, stdout, output_buffer, &output_start, &output_remaining);

        if ((error != NULL) && (error_remaining || (source == port_file_source(fileno(error)))))
            transfer(error, stderr, error_buffer, &error_start, &error_remaining);

        /* Handle process termination. */

        if (source == port_process_source(pid))
        {
            transfer(output, stdout, output_buffer, &output_start, &output_remaining);

            if (error != NULL)
                transfer(error, stderr, error_buffer, &error_start, &error_remaining);

            result = port_source_event_values(source);
            break;
        }
    }

    /* Close any open pipes. */

    if (input != NULL)
        fclose(input);

    fclose(output);

    if (error != NULL)
        fclose(error);

    printf("Result: %ld\n", result.val);
    return 0;
}

/* vim: tabstop=4 expandtab shiftwidth=4
*/
